
#include <SDL.h>
#include <GL/glew.h>
#include <GL/glu.h>

#include <stdio.h>
#include <math.h>

#include "video.h"
#include "timer.h"
#include "model.h"
#include "shadersource.h"

#include "common.h"

#include "render.h"

static GLuint shaderProgram, vertexShader, fragmentShader;
static GLint shaderTime,
  shaderMeshColor,
  shaderShadingEnable,
  shaderSheenEnable, shaderSheenColor,
  shaderZShorten;

static SDL_GLContext glcontext;

int Render_Init(void) {
  // gl context setup
  glcontext = SDL_GL_CreateContext(window);

  // enable vsync
  if(windowData.vsyncEnabled) SDL_GL_SetSwapInterval(1);

  // display opengl renderer version etc.
  const char *renderer = (const char *)glGetString(GL_RENDERER);
  const char *vendor   = (const char *)glGetString(GL_VENDOR);
  const char *version  = (const char *)glGetString(GL_VERSION);

  printf("Renderer: %s (%s)\n", renderer, vendor);
  printf("Driver: %s\n", version);

  // check extensions
  GLenum glewError = glewInit();
  if(glewError != GLEW_OK) {
    printf("Render_Init: Could not initialize GLEW: %s\n", glewGetErrorString(glewError));
    return 1;
  }

  if(!GLEW_ARB_shader_objects) {
    printf("Render_Init: Missing OpenGL extension: ARB_shader_objects\n");
    return 1;
  }

  if(!GLEW_ARB_vertex_shader) {
    printf("Render_Init: Missing OpenGL extension: ARB_vertex_shader\n");
    return 1;
  }

  if(!GLEW_ARB_fragment_shader) {
    printf("Render_Init: Missing OpenGL extension: ARB_fragment_shader\n");
    return 1;
  }

  // gl state setup
  glClearColor(0.2f, 0.2f, 0.2f, 1.0f);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // gl flags
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_BLEND);

  // shaders

  Shadersource_LoadVertexShader("data/shaders/main.vert");
  Shadersource_LoadFragmentShader("data/shaders/main.frag");

  // vertex shader compilation
  vertexShader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(vertexShader, 1, (const GLchar**) &vertexShaderSource, NULL);  
  glCompileShader(vertexShader);

  GLint shaderStatus;
  glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &shaderStatus);

  char shaderStatusString[512];
  glGetShaderInfoLog(vertexShader, 512, NULL, shaderStatusString);

  if(shaderStatus != GL_TRUE) {
    printf("Vertex shader compile status (Fail): %s\n", shaderStatusString);    
  } else {
    printf("Vertex shader compile status (Success): %s\n", shaderStatusString);
  }

  // fragment shader compilation
  fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fragmentShader, 1, (const GLchar**) &fragmentShaderSource, NULL);  
  glCompileShader(fragmentShader);

  glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &shaderStatus);

  glGetShaderInfoLog(fragmentShader, 512, NULL, shaderStatusString);

  if(shaderStatus != GL_TRUE) {
    printf("Fragment shader compile status (Fail): %s\n", shaderStatusString);    
  } else {
    printf("Fragment shader compile status (Success): %s\n", shaderStatusString);
  }

  // bind into a shader program
  shaderProgram = glCreateProgram();
  glAttachShader(shaderProgram, vertexShader);
  glAttachShader(shaderProgram, fragmentShader);

  glLinkProgram(shaderProgram);
  glUseProgram(shaderProgram);

  // set uniform locations
  shaderTime = glGetUniformLocation(shaderProgram, "time");

  shaderMeshColor = glGetUniformLocation(shaderProgram, "meshColor");

  shaderShadingEnable = glGetUniformLocation(shaderProgram, "shadingEnable");

  shaderSheenEnable = glGetUniformLocation(shaderProgram, "sheenEnable");
  shaderSheenColor = glGetUniformLocation(shaderProgram, "sheenColor");

  shaderZShorten = glGetUniformLocation(shaderProgram, "zShorten");

  return 0;
}

int Render_DeInit(void) {
  SDL_GL_DeleteContext(glcontext);
  glcontext = NULL;

  glDeleteShader(vertexShader);
  glDeleteShader(fragmentShader);

  glDeleteProgram(shaderProgram);

  return 0;
}

int Render_Frame(void) {
  // clear render buffers
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // viewport and perspective
  Render_SetPerspective(70.0f, 0.2f, 200.0f);

  // test rendering
  glPushMatrix();

  /*
  gluLookAt(0.0f, -5.0f, 0.0f,
	    0.0f, 1.0f, 0.0f,
	    0.0f, 0.0f, 1.0f);

  glTranslatef(0.0f, 0.0f, (GLfloat) sin(timerData.millis * 0.005f) * 0.15f);
  glRotatef((GLfloat) ((timerData.millis / 8) % 360), 0.0f, 0.0f, 1.0f);
  */

  /*
  gluLookAt(0.0, 0.0, 0.0,
	    0.0, 1.0, 0.0,
	    0.0, 0.0, 1.0);
  glTranslatef(0.2, 0.5, -0.3);
  glRotatef(180, 0.0, 0.0, 1.0);
  */

  glMatrixMode(GL_PROJECTION);
  gluLookAt(0.0f, 0.0f, 0.0f,
	    0.0f, 1.0f, 0.0f,
	    0.0f, 0.0f, 1.0f);
  glMatrixMode(GL_MODELVIEW);

  glUniform4f(shaderMeshColor, 1.0f, 1.0f, 1.0f, 1.0f);

  glUniform1i(shaderShadingEnable, 1);

  float viewbobz = sin(timerData.millis * 0.02f) * 0.004f;
  float jumpz = MIN(0, sin(timerData.millis * 0.006f) * 0.3);

  glPushMatrix();
  glTranslatef(0.0f, 2.0f - (timerData.millis * 0.002f), 0.2f + viewbobz + jumpz);
  glRotatef(-90, 0.0f, 0.0f, 1.0f);

  int i, j;

  // lever
  glUniform1i(shaderSheenEnable, 1);
  glUniform4f(shaderSheenColor, 0.0f, 0.6f, 1.0f, 0.3f);

  glPushMatrix();
  for(i = 0; i < 20; i++) {
    glPushMatrix();
    glTranslatef((GLfloat) -i, 0.5f, -0.2f);
    glCallList(Model_DisplayLists[MDL_LEVER][0]);
    glPopMatrix();
  }
  glPopMatrix();

  glUniform1i(shaderSheenEnable, 0);

  // wall
  glPushMatrix();
  for(i = 0; i < 20; i++) {
    glPushMatrix();
    glTranslatef((GLfloat) -i, 0.0f, 0.0f);
    glCallList(Model_DisplayLists[MDL_PREFAB_WALL_2][0]);
    glCallList(Model_DisplayLists[MDL_PREFAB_TRIM_1][0]);
    glPopMatrix();
  }
  glPopMatrix();

  // floor
  for(i = 0; i < 20; i++) {
    for(j = 0; j < 20; j++) {
      glPushMatrix();
      glTranslatef((GLfloat) -i, (GLfloat) -j, 0.0);
      glCallList(Model_DisplayLists[MDL_PREFAB_FLOOR_2][0]);
      glPopMatrix();
    }
  }
  glPopMatrix();

  // shotgun
  float idlebobz = sin(timerData.millis * 0.003f) * 0.015f;
  
  float bobspeed = 1.0f;

  float bobx = sin(timerData.millis * 0.01f * bobspeed) * 0.08f;
  float bobz = fabs(cos(timerData.millis * 0.01f * bobspeed) * 0.04f);
  float bobry = sin(timerData.millis * 0.02f * bobspeed) * 0.02f;
  
  glTranslatef(0.2f + bobx, 0.5f, -0.3f + bobz + idlebobz);
  glRotatef(180, 0.0f, bobry, 1.0f);
  
  glUniform1f(shaderTime, (GLfloat) (timerData.millis) / 1000.0f);

  glUniform1i(shaderZShorten, 1);
  //glCallList(Model_DisplayLists[MDL_SHOTGUN][(timerData.millis / 40) % 18]);
  glCallList(Model_DisplayLists[MDL_SHOTGUN][0]);
  glUniform1i(shaderZShorten, 0);
  
  glPopMatrix();
  // end test rendering

  return 0;
}

void Render_AfterFrame(void) {
  SDL_GL_SwapWindow(window);
}

void Render_SetViewport(int w, int h) {
  glViewport(0, 0, w, h);
}

void Render_SetPerspective(GLfloat fov, GLfloat clipNear, GLfloat clipFar) {
  double aspect = (GLfloat) windowData.width / (GLfloat) windowData.height;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective(fov, aspect, clipNear, clipFar);

  glMatrixMode(GL_MODELVIEW);
}
